import typing


class ArmorException(Exception):
    """Armor module basic exception"""

    def __init__(
        self,
        msg: typing.Optional[str] = None,
        stacktrace: typing.Optional[typing.Sequence[str]] = None,
    ):
        self.msg = msg
        self.stacktrace = stacktrace
        super().__init__()

    def __str__(self) -> str:
        exception_msg = f"Message: {self.msg}\n"
        if self.stacktrace:
            stacktrace = "\n".join(self.stacktrace)
            exception_msg += f"Stacktrace:\n{stacktrace}"
        return exception_msg


class ChallengeException(ArmorException):
    """hCAPTCHA Challenge basic exceptions"""


class ChallengeLangException(ChallengeException):
    """指定了不兼容的挑战语言"""


class ChallengePassed(ChallengeException):
    """挑战未弹出"""


class LoadImageTimeout(ChallengeException):
    """加载挑战图片超时"""


class ChallengeTimeout(ChallengeException):
    """人机挑战超时 CPU能力太弱无法在规定时间内完成挑战"""


class LabelNotFoundException(ChallengeException):
    """获取到空的图像标签名"""


class AssertTimeout(ChallengeTimeout):
    """断言超时"""



# -*- coding: utf-8 -*-
# Time       : 2022/2/15 17:40
# Author     : QIN2DIM
# Github     : https://github.com/QIN2DIM
# Description: 🚀 Yo Challenger!
from fire import Fire

from services.scaffold import Scaffold

if __name__ == "__main__":
    Fire(Scaffold)



# -*- coding: utf-8 -*-
# Time       : 2022-03-01 20:32
# Author     : Bingjie Yan
# Github     : https://github.com/beiyuouo
# Description:
import time
from typing import Optional

import cv2
import numpy as np
from skimage import feature
from skimage.future import graph
from skimage.segmentation import slic


class SKRecognition:
    @staticmethod
    def _weight_mean_color(graph_, src: int, dst: int, n: int):  # noqa
        """Callback to handle merging nodes by recomputing mean color.

        The method expects that the mean color of `dst` is already computed.

        Parameters
        ----------
        graph_ : RAG
            The graph under consideration.
        src, dst : int
            The vertices in `graph` to be merged.
        n : int
            A neighbor of `src` or `dst` or both.

        Returns
        -------
        data : dict
            A dictionary with the `"weight"` attribute set as the absolute
            difference of the mean color between node `dst` and `n`.
        """

        diff = graph_.nodes[dst]["mean color"] - graph_.nodes[n]["mean color"]
        diff = np.linalg.norm(diff)
        return {"weight": diff}

    @staticmethod
    def _merge_mean_color(graph_, src: int, dst: int):
        """Callback called before merging two nodes of a mean color distance graph.

        This method computes the mean color of `dst`.

        Parameters
        ----------
        graph_ : RAG
            The graph under consideration.
        src, dst : int
            The vertices in `graph` to be merged.
        """
        graph_.nodes[dst]["total color"] += graph_.nodes[src]["total color"]
        graph_.nodes[dst]["pixel count"] += graph_.nodes[src]["pixel count"]
        graph_.nodes[dst]["mean color"] = (
                graph_.nodes[dst]["total color"] / graph_.nodes[dst]["pixel count"]
        )

    @staticmethod
    def _remove_border(img):
        img[:, 1] = 0
        img[:, -2] = 0
        img[1, :] = 0
        img[-2, :] = 0
        return img

    def solution(self, img_stream, **kwargs) -> bool:
        """Implementation process of solution"""
        raise NotImplementedError


class _VerticalRiverRecognition(SKRecognition):
    """A fast solution for identifying vertical rivers"""

    def __init__(self, on_rainbow: Optional[str] = None):
        super().__init__(on_rainbow=on_rainbow)
        self.rainbow_key = "vertical river"

    def solution(self, img_stream, **kwargs) -> bool:
        """Implementation process of solution"""
        img_arr = np.frombuffer(img_stream, np.uint8)
        img = cv2.imdecode(img_arr, flags=1)

        img = cv2.pyrMeanShiftFiltering(img, sp=10, sr=40)
        img = cv2.bilateralFilter(img, d=9, sigmaColor=100, sigmaSpace=75)

        labels = slic(img, compactness=30, n_segments=400, start_label=1)
        g = graph.rag_mean_color(img, labels)

        labels2 = graph.merge_hierarchical(
            labels,
            g,
            thresh=35,
            rag_copy=False,
            in_place_merge=True,
            merge_func=self._merge_mean_color,
            weight_func=self._weight_mean_color,
        )

        return len(np.unique(labels2[-1])) >= 3


class _LeftPlaneRecognition(SKRecognition):
    """A fast solution for identifying `airplane in the sky flying left`"""

    def __init__(self, on_rainbow: Optional[str] = None):
        super().__init__(on_rainbow=on_rainbow)
        self.sky_threshold = 1800
        self.left_threshold = 30
        self.rainbow_key = "airplane in the sky flying left"

    def solution(self, img_stream: bytes, **kwargs) -> bool:
        """Implementation process of solution"""
        img_arr = np.frombuffer(img_stream, np.uint8)
        img = cv2.imdecode(img_arr, flags=1)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

        edges1 = feature.canny(img)
        edges1 = self._remove_border(edges1)

        # on the ground
        if np.count_nonzero(edges1) > self.sky_threshold:
            return False

        min_x = np.min(np.nonzero(edges1), axis=1)[1]
        max_x = np.max(np.nonzero(edges1), axis=1)[1]

        left_nonzero = np.count_nonzero(edges1[:, min_x: min(max_x, min_x + self.left_threshold)])
        right_nonzero = np.count_nonzero(edges1[:, max(min_x, max_x - self.left_threshold): max_x])

        # Flying towards the right
        if left_nonzero > right_nonzero:
            return False

        time.sleep(0.25)
        return True


class _RightPlaneRecognition(SKRecognition):
    def __init__(self, on_rainbow: Optional[str] = None):
        super().__init__(on_rainbow=on_rainbow)
        self.sky_threshold = 1800
        self.left_threshold = 30
        self.rainbow_key = "airplanes in the sky that are flying to the right"

    def solution(self, img_stream: bytes, **kwargs) -> bool:
        img_arr = np.frombuffer(img_stream, np.uint8)
        img = cv2.imdecode(img_arr, flags=1)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

        edges1 = feature.canny(img)
        edges1 = self._remove_border(edges1)

        # On the ground
        if np.count_nonzero(edges1) > self.sky_threshold:
            return False

        min_x = np.min(np.nonzero(edges1), axis=1)[1]
        max_x = np.max(np.nonzero(edges1), axis=1)[1]

        left_nonzero = np.count_nonzero(edges1[:, min_x: min(max_x, min_x + self.left_threshold)])
        right_nonzero = np.count_nonzero(edges1[:, max(min_x, max_x - self.left_threshold): max_x])

        # Flying towards the left
        if left_nonzero < right_nonzero:
            return False

        time.sleep(0.15)
        return True



# -*- coding: utf-8 -*-
# Time       : 2022/1/16 0:25
# Author     : QIN2DIM
# Github     : https://github.com/QIN2DIM
# Description:
import time
import typing
import warnings

from loguru import logger
from selenium.common.exceptions import WebDriverException

from services.hcaptcha_challenger import HolyChallenger
from services.hcaptcha_challenger.exceptions import ChallengePassed
from services.settings import HCAPTCHA_DEMO_SITES, DIR_MODEL, DIR_CHALLENGE, PATH_OBJECTS_YAML
from services.utils import get_challenge_ctx

warnings.filterwarnings("ignore", category=DeprecationWarning)


@logger.catch()
def test():
    """Check if the Challenger driver version is compatible"""
    ctx = get_challenge_ctx(silence=True)
    try:
        ctx.get(HCAPTCHA_DEMO_SITES[0])
    finally:
        ctx.quit()
    logger.success("The adaptation is successful")


def _motion(sample_site: str, ctx, challenger: HolyChallenger) -> typing.Optional[str]:
    resp = None

    # Read the hCaptcha challenge test site
    ctx.get(sample_site)

    # Detects if a clickable `hcaptcha checkbox` appears on the current page.
    # The `sample site` must pop up the `checkbox`, where the flexible wait time defaults to 5s.
    # If the `checkbox` does not load in 5s, your network is in a bad state.
    if challenger.utils.face_the_checkbox(ctx):
        # Enter iframe-checkbox --> Process hcaptcha checkbox --> Exit iframe-checkbox
        challenger.anti_checkbox(ctx)
        # Enter iframe-content --> process hcaptcha challenge --> exit iframe-content
        resp = challenger.anti_hcaptcha(ctx)
    return resp


@logger.catch()
def runner(
    sample_site: str,
    lang: typing.Optional[str] = "zh",
    silence: typing.Optional[bool] = False,
    onnx_prefix: typing.Optional[str] = None,
    screenshot: typing.Optional[bool] = False,
    repeat: typing.Optional[int] = 10,
):
    """Human-Machine Challenge Demonstration | Top Interface"""

    # Instantiating Challenger Components
    challenger = HolyChallenger(
        dir_workspace=DIR_CHALLENGE,
        dir_model=DIR_MODEL,
        path_objects_yaml=PATH_OBJECTS_YAML,
        onnx_prefix=onnx_prefix,
        screenshot=screenshot,
        lang=lang,
        debug=True,
    )

    # Instantiating the Challenger Drive
    with get_challenge_ctx(silence=silence, lang=lang) as ctx:
        for i in range(repeat):
            start = time.time()
            try:
                if (resp := _motion(sample_site, ctx=ctx, challenger=challenger)) is None:
                    logger.warning("UnknownMistake")
                elif resp == challenger.CHALLENGE_SUCCESS:
                    challenger.log(f"End of demo - total: {round(time.time() - start, 2)}s")
                    logger.success(f"PASS[{i + 1}|{repeat}]".center(28, "="))
                elif resp == challenger.CHALLENGE_RETRY:
                    ctx.refresh()
                    logger.error(f"RETRY[{i + 1}|{repeat}]".center(28, "="))
            except ChallengePassed:
                ctx.refresh()
                logger.success(f"PASS[{i + 1}|{repeat}]".center(28, "="))
            except WebDriverException as err:
                logger.warning(err)



# -*- coding: utf-8 -*-
# Time       : 2022/4/30 22:34
# Author     : QIN2DIM
# Github     : https://github.com/QIN2DIM
# Description:
import json
import os
import shutil
import time
import typing
from os.path import join
from urllib.request import getproxies

import cv2
import requests
from loguru import logger


class ChallengeStyle:
    WATERMARK = 100
    GENERAL = 128
    GAN = 144


class Memory:
    _fn2memory = {}

    ASSET_TOKEN = "RA_kw"

    def __init__(self, fn: str, dir_memory: str = None):
        self.fn = fn
        self._dir_memory = "model/_memory" if dir_memory is None else dir_memory

        self._build()

    def _build(self) -> typing.Optional[typing.Dict[str, str]]:
        """:return filename to nodeId"""
        if not self._fn2memory:
            os.makedirs(self._dir_memory, exist_ok=True)
            for memory_name in os.listdir(self._dir_memory):
                fn = memory_name.split(".")[0]
                fn = fn if fn.endswith(".onnx") else f"{fn}.onnx"
                node_id = memory_name.split(".")[-1]
                if node_id.startswith(self.ASSET_TOKEN):
                    self._fn2memory[fn] = node_id
        return self._fn2memory

    def get_node_id(self) -> typing.Optional[str]:
        return self._fn2memory.get(self.fn, "")

    def dump(self, new_node_id: str):
        old_node_id = self._fn2memory.get(self.fn)
        self._fn2memory[self.fn] = new_node_id

        if not old_node_id:
            memory_name = join(self._dir_memory, f"{self.fn}.{new_node_id}")
            with open(memory_name, "w", encoding="utf8") as file:
                file.write(memory_name)
        else:
            memory_src = join(self._dir_memory, f"{self.fn}.{old_node_id}")
            memory_dst = join(self._dir_memory, f"{self.fn}.{new_node_id}")
            shutil.move(memory_src, memory_dst)

    def is_outdated(self, remote_node_id: str) -> typing.Optional[bool]:
        """延迟反射的诊断步骤，保持分布式网络的两端一致性"""
        local_node_id = self.get_node_id()

        # Invalid judgment
        if (
            not local_node_id
            or not remote_node_id
            or not isinstance(remote_node_id, str)
            or not remote_node_id.startswith(self.ASSET_TOKEN)
        ):
            return

        if local_node_id != remote_node_id:
            return True
        return False


class Assets:
    GITHUB_RELEASE_API = "https://api.github.com/repos/qin2dim/hcaptcha-challenger/releases"

    NAME_ASSETS = "assets"
    NAME_ASSET_NAME = "name"
    NAME_ASSET_SIZE = "size"
    NAME_ASSET_DOWNLOAD_URL = "browser_download_url"
    NAME_ASSET_NODE_ID = "node_id"

    _fn2assets = {}

    # Cache validity period: 2h
    CACHE_CONTROL = 7200

    def __init__(self, fn: str, dir_assets: str = None):
        self.fn = fn
        self._dir_assets = "model/_assets" if dir_assets is None else dir_assets

        self._pull()

    def _preload(self):
        """仅在 Assets._pull 网络请求发起前调用，将有效的本地缓存代替远端资源"""
        if self._fn2assets:
            return

        os.makedirs(self._dir_assets, exist_ok=True)
        assets = [i for i in os.listdir(self._dir_assets) if i.isdigit()]
        if len(assets) >= 1:
            asset_name = assets[-1]
            if int(asset_name) + self.CACHE_CONTROL > int(time.time()):
                recoded_name = join(self._dir_assets, asset_name)
                try:
                    with open(recoded_name, "r", encoding="utf8") as file:
                        self._fn2assets = json.load(file)
                except json.decoder.JSONDecodeError as err:
                    logger.warning(err)

    def _offload(self):
        """仅在 Assets._pull 网络请求发起后实现，用于更新本地缓存的内容及时间戳"""
        os.makedirs(self._dir_assets, exist_ok=True)
        for asset_fn in os.listdir(self._dir_assets):
            asset_src = join(self._dir_assets, asset_fn)
            asset_dst = join(self._dir_assets, f"_{asset_fn}")
            shutil.move(asset_src, asset_dst)
        recoded_name = join(self._dir_assets, str(int(time.time())))
        with open(recoded_name, "w", encoding="utf8") as file:
            json.dump(self._fn2assets, file)

    def _pull(self, skip_preload: bool = False) -> typing.Optional[typing.Dict[str, dict]]:
        def request_assets():
            logger.debug(f"Pulling AssetsObject from {self.GITHUB_RELEASE_API}")

            try:
                session = requests.session()
                resp = session.get(self.GITHUB_RELEASE_API, proxies=getproxies(), timeout=3)
                data = resp.json()[0]
            except (requests.exceptions.ConnectionError, json.decoder.JSONDecodeError) as err:
                logger.error(err)
            except (AttributeError, IndexError, KeyError) as err:
                logger.error(err)
            else:
                if isinstance(data, dict):
                    assets: typing.List[dict] = data.get(self.NAME_ASSETS, [])
                    for asset in assets:
                        self._fn2assets[asset[self.NAME_ASSET_NAME]] = asset
            finally:
                self._offload()

        if not skip_preload:
            self._preload()
        if not self._fn2assets:
            request_assets()
        return self._fn2assets

    def _get_asset(self, key: str, oncall_default: typing.Any):
        return self._fn2assets.get(self.fn, {}).get(key, oncall_default)

    def sync(self, force: typing.Optional[bool] = None, **kwargs):
        raise NotImplementedError

    @property
    def dir_assets(self):
        return self._dir_assets

    def get_node_id(self) -> typing.Optional[str]:
        return self._get_asset(self.NAME_ASSET_NODE_ID, "")

    def get_download_url(self) -> typing.Optional[str]:
        return self._get_asset(self.NAME_ASSET_DOWNLOAD_URL, "")

    def get_size(self) -> typing.Optional[int]:
        return self._get_asset(self.NAME_ASSET_SIZE, 0)


class PluggableObjects:
    URL_REMOTE_OBJECTS = (
        "https://raw.githubusercontent.com/QIN2DIM/hcaptcha-challenger/main/src/objects.yaml"
    )

    DEFAULT_FILENAME = "objects.yaml"

    def __init__(self, path_objects: str):
        self.path_objects = path_objects
        if not os.path.isfile(self.path_objects):
            self.path_objects = self.DEFAULT_FILENAME
        self.fn = os.path.basename(self.path_objects)

    def sync(self):
        _request_asset(self.URL_REMOTE_OBJECTS, self.path_objects, self.fn)


class ModelHub:
    _fn2net = {}

    def __init__(self, onnx_prefix: str, name: str, dir_model: str):
        """
        :param onnx_prefix: 模型文件名，不含有 ext
        :param name: 日志打印显示的标记
        :param dir_model: 模型所在的本地目录
        """
        self._dir_model = "model" if dir_model is None else dir_model

        self.net = None
        self.flag = name
        self.fn = f"{onnx_prefix}.onnx" if not onnx_prefix.endswith(".onnx") else onnx_prefix
        self.path_model = join(dir_model, self.fn)

        self.memory = Memory(fn=self.fn, dir_memory=join(dir_model, "_memory"))
        self.assets = Assets(fn=self.fn, dir_assets=join(dir_model, "_assets"))

    @logger.catch()
    def pull_model(self, fn: str = None, path_model: str = None):
        """
        1. node_id: Record the insertion point
        and indirectly judge the changes of the file with the same name

        2. assets.List: Record the item list of the release attachment,
        and directly determine whether there are undownloaded files

        3. assets.size: Record the amount of bytes inserted into the file,
        and directly determine whether the file is downloaded completely
        :param fn:
        :param path_model:
        :return:
        """
        fn = self.fn if fn is None else fn
        path_model = self.path_model if path_model is None else path_model

        asset_node_id = self.assets.get_node_id()
        asset_download_url = self.assets.get_download_url()
        asset_size = self.assets.get_size()

        # Check for extreme cases
        if (
            not fn.endswith(".onnx")
            or not isinstance(asset_download_url, str)
            or not asset_download_url.startswith("https:")
        ):
            return

        # Matching conditions to trigger download tasks
        if (
            not os.path.exists(path_model)
            or os.path.getsize(path_model) != asset_size
            or self.memory.is_outdated(remote_node_id=asset_node_id)
        ):
            _request_asset(asset_download_url, path_model, fn)
            self.memory.dump(new_node_id=asset_node_id)

    @logger.catch()
    def register_model(self) -> typing.Optional[bool]:
        """Load and register an existing model"""
        # Update AssetsObject local cache
        if os.path.exists(self.path_model) and not self.memory.is_outdated(
            self.assets.get_node_id()
        ):
            self.net = cv2.dnn.readNetFromONNX(self.path_model)
            self._fn2net[self.fn] = self.net
            return True
        return False

    def offload(self):
        if self.fn in self._fn2net:
            self._fn2net.pop(self.fn)

    def match_net(self):
        """
        PluggableONNXModel 对象实例化时：
        - 自动读取并注册 objects.yaml 中注明的且已存在指定目录的模型对象，
        - 然而，objects.yaml 中表达的标签组所对应的模型文件不一定都已存在。
        - 初始化时不包含新的网络请求，即，不在初始化阶段下载缺失模型。

        match_net 模型被动拉取：
        - 在挑战进行时被动下载缺失的用于处理特定二分类任务的 ONNX 模型。
        - 匹配的模型会被自动下载、注册并返回。
        - 不在 objects.yaml 名单中的模型不会被下载

        升级的模型不支持热加载，需要重启程序才能读入，但新插入的模型可直接使用。
        :return:
        """
        if not self.net:
            self.pull_model()
            self.register_model()
        return self.net

    @property
    def fn2net(self):
        """check model objs"""
        return self._fn2net

    def solution(self, img_stream, **kwargs) -> bool:
        """Implementation process of solution"""
        raise NotImplementedError


def _request_asset(asset_download_url: str, asset_path: str, fn_tag: str):
    headers = {
        "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) "
        "Chrome/105.0.0.0 Safari/537.36 Edg/105.0.1343.27"
    }
    logger.debug(f"Downloading {fn_tag} from {asset_download_url}")

    # FIXME: PTC-W6004
    #  Audit required: External control of file name or path
    with open(asset_path, "wb") as file, requests.get(
        asset_download_url, headers=headers, stream=True, proxies=getproxies()
    ) as response:
        for chunk in response.iter_content(chunk_size=1024):
            if chunk:
                file.write(chunk)



# -*- coding: utf-8 -*-
# Time       : 2022/7/20 5:41
# Author     : QIN2DIM
# Github     : https://github.com/QIN2DIM
# Description:
import csv
import os.path
import time
from datetime import datetime

from loguru import logger
from selenium.common.exceptions import WebDriverException
from selenium.webdriver import Chrome, ChromeOptions
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from webdriver_manager.chrome import ChromeDriverManager

from services.utils import ToolBox


class MotionData:
    def __init__(self, dir_database: str = None):
        self.ctx_session = None
        self.dir_log = "tracker" if dir_database is None else dir_database

        self.action_name = "MotionData"
        self.startpoint = time.time()
        self.sequential_queue = {}

    def __enter__(self):
        options = ChromeOptions()
        service = Service(ChromeDriverManager().install())
        self.ctx_session = Chrome(service=service, options=options)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        logger.debug(
            ToolBox.runtime_report(
                motive="QUIT",
                action_name=self.action_name,
                message="Turn off tracker",
                runtime=f"{round(time.time() - self.startpoint, 2)}s",
            )
        )

        try:
            if self.ctx_session:
                self._overload(self.ctx_session)
                self._offload()
                self.ctx_session.quit()
        except AttributeError:
            pass

    def _overload(self, ctx):
        try:
            mouse_track_obj = ctx.find_element(By.CLASS_NAME, "track-coordinate-list")
            mouse_track = mouse_track_obj.text
        except (WebDriverException, AttributeError):
            logger.warning("Failed to record mouse track")
        except Exception as err:
            logger.debug(err)
        else:
            if not mouse_track:
                return
            for p in mouse_track.split(","):
                x = [int(xi.split(".")[0]) for xi in p.split(":")]
                self.sequential_queue[x[0]] = [x[1], x[2]]

    def _offload(self):
        endpoint = (
            str(datetime.utcnow()).replace("-", "").replace(":", "").replace(" ", "").split(".")[0]
        )

        # Record track
        fn = os.path.join(self.dir_log, "motion_data", f"{endpoint}.csv")
        os.makedirs(os.path.dirname(fn), exist_ok=True)
        with open(fn, "w", encoding="utf8", newline="") as file:
            writer = csv.writer(file)
            for xp, yp in self.sequential_queue.values():
                writer.writerow([xp, yp])

        # Record offset
        fn_offset = os.path.join(os.path.dirname(fn), f"{endpoint}.offset.csv")
        with open(fn_offset, "w", encoding="utf8", newline="") as file:
            writer = csv.writer(file)
            timeline = list(self.sequential_queue)[200:-200]
            for i in range(len(timeline) - 1):
                x1, y1 = self.sequential_queue[timeline[i]]
                x2, y2 = self.sequential_queue[timeline[i + 1]]
                writer.writerow([x2 - x1, y2 - y1])

        logger.success(
            ToolBox.runtime_report(
                motive="OFFLOAD",
                action_name=self.action_name,
                message="Record mouse track",
                endpoint=endpoint,
                path=fn,
            )
        )

    def mimic(self, url: str = "http://127.0.0.1:8000"):
        if not self.ctx_session:
            return

        try:
            self.ctx_session.get(url)
            logger.debug("Press CTRL + C to terminate the action")
            for _ in range(120):
                time.sleep(0.24)
                self._overload(self.ctx_session)
        except (KeyboardInterrupt, EOFError):
            logger.debug("Received keyboard interrupt signal")



# -*- coding: utf-8 -*-
# Time       : 2021/12/22 9:05
# Author     : QIN2DIM
# Github     : https://github.com/QIN2DIM
# Description:
import asyncio
import typing

import aiohttp


class AshFramework:
    """轻量化的协程控件"""

    def __init__(self, docker: typing.Optional[typing.List] = None):
        # 任务容器：queue
        self.worker, self.done = asyncio.Queue(), asyncio.Queue()
        # 任务容器
        self.docker = docker
        # 任务队列满载时刻长度
        self.max_queue_size = 0

    def progress(self) -> str:
        """任务进度"""
        _progress = self.max_queue_size - self.worker.qsize()
        return f"{_progress}/{self.max_queue_size}"

    def preload(self):
        """预处理"""

    def overload(self):
        """任务重载"""
        if self.docker:
            for task in self.docker:
                self.worker.put_nowait(task)
        self.max_queue_size = self.worker.qsize()

    def offload(self) -> typing.Optional[typing.List]:
        """缓存卸载"""
        crash = []
        while not self.done.empty():
            crash.append(self.done.get())
        return crash

    async def control_driver(self, context, session=None):
        """需要并发执行的代码片段"""
        raise NotImplementedError

    async def launcher(self, session=None):
        """适配接口模式"""
        while not self.worker.empty():
            context = self.worker.get_nowait()
            await self.control_driver(context, session=session)

    async def subvert(self, workers: typing.Union[str, int]):
        """
        框架接口

        loop = asyncio.get_event_loop()
        loop.run_until_complete(fl.go(workers))

        :param workers: ["fast", power]
        :return:
        """
        # 任务重载
        self.overload()

        # 弹出空载任务
        if self.max_queue_size == 0:
            return

        # 粘性功率
        workers = self.max_queue_size if workers in ["fast"] else workers
        workers = workers if workers <= self.max_queue_size else self.max_queue_size

        # 弹性分发
        task_list = []
        async with aiohttp.ClientSession() as session:
            for _ in range(workers):
                task = asyncio.create_task(self.launcher(session=session))
                task_list.append(task)
            await asyncio.wait(task_list)



# -*- coding: utf-8 -*-
# Time       : 2022/1/20 16:16
# Author     : QIN2DIM
# Github     : https://github.com/QIN2DIM
# Description:
import shutil
import sys
import typing
import webbrowser

from loguru import logger
from webdriver_manager.chrome import ChromeType
from webdriver_manager.core.utils import get_browser_version_from_os

from services.hcaptcha_challenger import YOLO, PluggableObjects
from services.settings import DIR_MODEL, DIR_ASSETS


def download_driver():
    # Detect environment variable `google-chrome`.
    browser_version = get_browser_version_from_os(ChromeType.GOOGLE)
    if browser_version != "UNKNOWN":
        return

    # `google-chrome` is missing from environment variables, prompting players to install manually.
    logger.critical(
        "The current environment variable is missing `google-chrome`, "
        "please install Chrome for your system"
    )
    logger.info(
        "Ubuntu: https://linuxize.com/post/how-to-install-google-chrome-web-browser-on-ubuntu-20-04/"
    )
    logger.info(
        "CentOS 7/8: https://linuxize.com/post/how-to-install-google-chrome-web-browser-on-centos-7/"
    )
    if "linux" not in sys.platform:
        webbrowser.open("https://www.google.com/chrome/")

    logger.info("Re-execute the `install` scaffolding command after the installation is complete.")


def do(yolo_onnx_prefix: typing.Optional[str] = None, upgrade: typing.Optional[bool] = False):
    """下载项目运行所需的各项依赖"""
    dir_assets = DIR_ASSETS

    download_driver()

    if upgrade is True:
        logger.debug(f"Reloading the local cache of Assets {dir_assets}")
        shutil.rmtree(dir_assets, ignore_errors=True)
    PluggableObjects(dir_assets).sync()

    # PULL YOLO ONNX Model by the prefix flag
    YOLO(DIR_MODEL, yolo_onnx_prefix).pull_model()


def summon(
    dir_assets: str,
    dir_model: str,
    onnx_prefix: typing.Optional[str] = "yolov6n",
    upgrade: typing.Optional[bool] = False,
):
    if upgrade is True:
        logger.debug(f"Reloading the local cache of Assets {dir_assets}")
        shutil.rmtree(dir_assets, ignore_errors=True)
    PluggableObjects(dir_assets).sync()
    YOLO(dir_model, onnx_prefix=onnx_prefix).pull_model()



# -*- coding: utf-8 -*-
# Time       : 2022/1/16 0:25
# Author     : QIN2DIM
# Github     : https://github.com/QIN2DIM
# Description:
from .accelerator.core import AshFramework
from .toolbox.toolbox import ToolBox, get_challenge_ctx

__all__ = ["ToolBox", "AshFramework", "get_challenge_ctx"]



# -*- coding: utf-8 -*-
# Time       : 2022/2/15 17:43
# Author     : QIN2DIM
# Github     : https://github.com/QIN2DIM
# Description:
from .core import HolyChallenger, ArmorUtils
from .solutions.kernel import PluggableObjects
from .solutions.resnet import PluggableONNXModels
from .solutions.yolo import YOLO

__all__ = ["YOLO", "HolyChallenger", "ArmorUtils", "PluggableONNXModels", "PluggableObjects"]



# -*- coding: utf-8 -*-
# Time       : 2022/2/15 17:42
# Author     : QIN2DIM
# Github     : https://github.com/QIN2DIM
# Description:
import os
from os.path import join, dirname

from services.utils import ToolBox

HCAPTCHA_DEMO_API = "https://accounts.hcaptcha.com/demo?sitekey={}"
_SITE_KEYS = {
    "epic": "91e4137f-95af-4bc9-97af-cdcedce21c8c",
    "hcaptcha": "a5f74b19-9e45-40e0-b45d-47ff91b7a6c2",
    "discord": "f5561ba9-8f1e-40ca-9b5b-a0b3f719ef34",
    "oracle": "d857545c-9806-4f9e-8e9d-327f565aeb46",
    "publisher": "c86d730b-300a-444c-a8c5-5312e7a93628",
}

# https://www.wappalyzer.com/technologies/security/hcaptcha/
HCAPTCHA_DEMO_SITES = [
    # [√] label: Tags follow point-in-time changes
    HCAPTCHA_DEMO_API.format(_SITE_KEYS["publisher"]),
    # [√] label: `vertical river`
    HCAPTCHA_DEMO_API.format(_SITE_KEYS["oracle"]),
    # [x] label: `airplane in the sky flying left`
    HCAPTCHA_DEMO_API.format(_SITE_KEYS["discord"]),
    # [√] label: hcaptcha-challenger
    HCAPTCHA_DEMO_API.format(_SITE_KEYS["hcaptcha"]),
]

# ---------------------------------------------------
# [√]Lock the project directory
# ---------------------------------------------------
# hcaptcha-challenger
#  ├── database
#  │   ├── logs
#  │   ├── motion_data
#  │   └── temp_cache
#  │       ├── _challenge
#  │       └── captcha_screenshot
#  ├── model
#  │   ├── _assets
#  │   ├── _memory
#  │   └── rainbow.yaml[DEPRECATED]
#  └── src
#      ├── main.py
#      └── objects.yaml
# ---------------------------------------------------
PROJECT_SRC = dirname(dirname(__file__))

DIR_DATABASE = join(dirname(PROJECT_SRC), "database")
DIR_LOG = join(DIR_DATABASE, "logs")
DIR_TEMP_CACHE = join(DIR_DATABASE, "temp_cache")
DIR_CHALLENGE = join(DIR_TEMP_CACHE, "_challenge")

DIR_MODEL = join(dirname(PROJECT_SRC), "model")
DIR_ASSETS = join(DIR_MODEL, "_assets")

PATH_OBJECTS_YAML = join(PROJECT_SRC, "objects.yaml")

# ---------------------------------------------------
# [√]Server log configuration
# ---------------------------------------------------
logger = ToolBox.init_log(error=join(DIR_LOG, "error.log"), runtime=join(DIR_LOG, "runtime.log"))

# ---------------------------------------------------
# [√]Path completion
# ---------------------------------------------------
for _pending in (DIR_CHALLENGE, DIR_ASSETS):
    os.makedirs(_pending, exist_ok=True)



# -*- coding: utf-8 -*-
# Time       : 2022/2/15 17:54
# Author     : QIN2DIM
# Github     : https://github.com/QIN2DIM
# Description:
from services import settings
from services.hcaptcha_challenger import HolyChallenger
from services.hcaptcha_challenger import exceptions
from services.settings import logger
from services.utils import ToolBox, get_challenge_ctx
from .scaffold import install

__all__ = [
    "exceptions",
    "HolyChallenger",
    "ToolBox",
    "get_challenge_ctx",
    "settings",
    "logger",
    "install",
]



# -*- coding: utf-8 -*-
# Time       : 2022/7/20 4:45
# Author     : QIN2DIM
# Github     : https://github.com/QIN2DIM
# Description:
import os.path

from sanic import Sanic, Request
from sanic.response import html

app = Sanic("motion-data")


@app.route("/")
async def test(request: Request):
    print(request.headers)
    fp = os.path.join(os.path.dirname(__file__), "motion.html")
    with open(fp, "rb") as file:
        return html(file.read())



# -*- coding: utf-8 -*-
# Time       : 2022/7/20 5:46
# Author     : QIN2DIM
# Github     : https://github.com/QIN2DIM
# Description:
from services.motion_data.offload import MotionData
from services.settings import DIR_DATABASE


def train_motion(test_site: str):
    with MotionData(DIR_DATABASE) as motion:
        motion.mimic(test_site)



# -*- coding: utf-8 -*-
# Time       : 2022/1/16 0:27
# Author     : QIN2DIM
# Github     : https://github.com/QIN2DIM
# Description:
import logging
import os
import sys
import typing

import undetected_chromedriver as uc
from loguru import logger
from selenium.common.exceptions import WebDriverException
from webdriver_manager.chrome import ChromeDriverManager, ChromeType
from webdriver_manager.core.utils import get_browser_version_from_os


class ToolBox:
    """Portable Toolbox"""

    @staticmethod
    def init_log(**sink_path):
        """Initialize loguru log information"""
        event_logger_format = (
            "<g>{time:YYYY-MM-DD HH:mm:ss}</g> | "
            "<lvl>{level}</lvl> - "
            # "<c><u>{name}</u></c> | "
            "{message}"
        )
        logger.remove()
        logger.add(
            sink=sys.stdout,
            colorize=True,
            level="DEBUG",
            format=event_logger_format,
            diagnose=False,
        )
        if sink_path.get("error"):
            logger.add(
                sink=sink_path.get("error"),
                level="ERROR",
                rotation="1 week",
                encoding="utf8",
                diagnose=False,
            )
        if sink_path.get("runtime"):
            logger.add(
                sink=sink_path.get("runtime"),
                level="DEBUG",
                rotation="20 MB",
                retention="20 days",
                encoding="utf8",
                diagnose=False,
            )
        return logger

    @staticmethod
    def runtime_report(action_name: str, motive: str = "RUN", message: str = "", **params) -> str:
        """格式化输出"""
        flag_ = f">> {motive} [{action_name}]"
        if message != "":
            flag_ += f" {message}"
        if params:
            flag_ += " - "
            flag_ += " ".join([f"{i[0]}={i[1]}" for i in params.items()])

        return flag_


def get_challenge_ctx(silence: typing.Optional[bool] = None, lang: typing.Optional[str] = None):
    """
    Challenger drive for handling human-machine challenges.

    :param silence: Control headless browser
    :param lang: Restrict the language of hCAPTCHA label.
      See https://github.com/QIN2DIM/hcaptcha-challenger/issues/13
    :rtype: uc.Chrome
    """
    # Control headless browser
    silence = True if silence is None or "linux" in sys.platform else silence

    # - Restrict browser startup parameters
    options = uc.ChromeOptions()
    options.add_argument("--log-level=3")
    options.add_argument("--disable-dev-shm-usage")

    # - Restrict the language of hCaptcha label
    # - Environment variables are valid only in the current process
    # and do not affect other processes in the operating system
    os.environ["LANGUAGE"] = "zh_CN" if lang is None else lang
    options.add_argument(f"--lang={os.getenv('LANGUAGE')}")

    if silence is True:
        options.add_argument("--disable-gpu")
        options.add_argument("--disable-software-rasterizer")

    # - Use chromedriver cache to improve application startup speed
    # - Requirement: undetected-chromedriver >= 3.1.5.post2
    logging.getLogger("WDM").setLevel(logging.NOTSET)
    driver_executable_path = ChromeDriverManager().install()
    version_main = get_browser_version_from_os(ChromeType.GOOGLE).split(".")[0]

    logger.debug("🎮 Activate challenger context")
    try:
        return uc.Chrome(
            options=options, headless=silence, driver_executable_path=driver_executable_path
        )
    except WebDriverException:
        return uc.Chrome(
            options=options,
            headless=silence,
            version_main=int(version_main) if version_main.isdigit() else None,
        )



# -*- coding: utf-8 -*-
# Time       : 2022/1/16 0:25
# Author     : QIN2DIM
# Github     : https://github.com/QIN2DIM
# Description:
import typing

from apis.scaffold import install, challenge, app, motion
from services.settings import HCAPTCHA_DEMO_SITES, _SITE_KEYS, HCAPTCHA_DEMO_API


class Scaffold:
    """System scaffolding Top-level interface commands"""

    CHALLENGE_LANGUAGE = "zh"

    def __init__(self, lang: typing.Optional[str] = None):
        if lang is not None:
            Scaffold.CHALLENGE_LANGUAGE = lang

        self.challenge = self.demo
        self.run = self.demo

    @staticmethod
    def install(model: typing.Optional[str] = None, upgrade: typing.Optional[bool] = False):
        """Download Project Dependencies and upgrade all pluggable ONNX model"""
        install.do(yolo_onnx_prefix=model, upgrade=upgrade)

    @staticmethod
    def test():
        """Test the Challenger drive for fitment"""
        challenge.test()

    @staticmethod
    def demo(
        silence: typing.Optional[bool] = False,
        model: typing.Optional[str] = None,
        target: typing.Optional[str] = None,
        sitekey: typing.Optional[str] = None,
        screenshot: typing.Optional[bool] = False,
        repeat: typing.Optional[int] = 5,
    ):
        """
        Dueling with hCAPTCHA challenge using YOLOv5.

        Usage: python main.py demo [flags]
        ——————————————————————————————————————————————————————————————————
        or: python main.py demo --model=yolov5n6
        or: python main.py demo --target=discord
        or: python main.py demo --lang=en
        or: python main.py demo --sitekey=[UUID]
        ——————————————————————————————————————————————————————————————————

        :param repeat: Default 5. Number of times to repeat the presentation.
        :param screenshot: Default False. Save screenshot of the challenge result.
            PATH: database/temp_cache/captcha_screenshot/
            FILENAME: ChallengeLabelName.png
        :param sitekey: Default None. customize the challenge theme via sitekey
        :param silence: Default False. Whether to startup headless browser.
        :param model: Default "yolov5s6". Select the YOLO model to handle specific challenges.
            within [yolov5n6 yolov5s6 yolov5m6 yolov6n yolov6t yolov6s]
        :param target: Default None. Designate `Challenge Source`. See the global value SITE_KEYS.
        :return:
        """
        # Generate challenge topics
        if _SITE_KEYS.get(target):
            sample_site = HCAPTCHA_DEMO_API.format(_SITE_KEYS[target])
        else:
            sample_site = HCAPTCHA_DEMO_SITES[0]
        if sitekey is not None:
            sample_site = HCAPTCHA_DEMO_API.format(sitekey.strip())

        challenge.runner(
            sample_site,
            lang=Scaffold.CHALLENGE_LANGUAGE,
            silence=silence,
            onnx_prefix=model,
            screenshot=screenshot,
            repeat=repeat,
        )

    @staticmethod
    def tracker():
        app.run(debug=True, access_log=False)

    @staticmethod
    def motion():
        motion.train_motion("http://127.0.0.1:8000")
